package com.gmail.nossr50.datatypes.skills;

import com.gmail.nossr50.mcMMO;
import com.gmail.nossr50.config.Config;
import com.gmail.nossr50.locale.LocaleLoader;
import com.gmail.nossr50.skills.SkillManager;
import com.gmail.nossr50.skills.acrobatics.AcrobaticsManager;
import com.gmail.nossr50.skills.archery.ArcheryManager;
import com.gmail.nossr50.skills.axes.AxesManager;
import com.gmail.nossr50.skills.excavation.ExcavationManager;
import com.gmail.nossr50.skills.fishing.FishingManager;
import com.gmail.nossr50.skills.herbalism.HerbalismManager;
import com.gmail.nossr50.skills.mining.MiningManager;
import com.gmail.nossr50.skills.smelting.SmeltingManager;
import com.gmail.nossr50.skills.swords.SwordsManager;
import com.gmail.nossr50.skills.taming.TamingManager;
import com.gmail.nossr50.skills.unarmed.UnarmedManager;
import com.gmail.nossr50.util.StringUtils;

public enum SkillType {
    ACROBATICS(AcrobaticsManager.class),
    ARCHERY(ArcheryManager.class),
    AXES(AxesManager.class, AbilityType.SKULL_SPLITTER, ToolType.AXE),
    EXCAVATION(ExcavationManager.class, AbilityType.GIGA_DRILL_BREAKER, ToolType.SHOVEL),
    FISHING(FishingManager.class),
    HERBALISM(HerbalismManager.class, AbilityType.GREEN_TERRA, ToolType.HOE),
    MINING(MiningManager.class, AbilityType.SUPER_BREAKER, ToolType.PICKAXE),
    REPAIR(null), // TODO: Create a proper RepairManager class
    SMELTING(SmeltingManager.class),
    SWORDS(SwordsManager.class, AbilityType.SERRATED_STRIKES, ToolType.SWORD),
    TAMING(TamingManager.class),
    UNARMED(UnarmedManager.class, AbilityType.BERSERK, ToolType.FISTS),
    WOODCUTTING(null, AbilityType.TREE_FELLER, ToolType.AXE); // TODO: Create a proper WoodcuttingManager class

    private Class<? extends SkillManager> managerClass;
    private AbilityType abilityType;
    private ToolType toolType;

    private SkillType(Class<? extends SkillManager> managerClass) {
        this.managerClass = managerClass;
        abilityType = null;
        toolType = null;
    }

    private SkillType(Class<? extends SkillManager> managerClass, AbilityType abilityType, ToolType toolType) {
        this.managerClass = managerClass;
        this.abilityType = abilityType;
        this.toolType = toolType;
    }

    public Class<? extends SkillManager> getManagerClass() {
        return managerClass;
    }

    public AbilityType getAbilityType() {
        return abilityType;
    }

    /**
     * Get the max level of this skill.
     *
     * @return the max level of this skill
     */
    public int getMaxLevel() {
        return Config.getInstance().getLevelCap(this);
    }

    public boolean getPVPEnabled() {
        return Config.getInstance().getPVPEnabled(this);
    }

    public boolean getPVEEnabled() {
        return Config.getInstance().getPVEEnabled(this);
    }

    public boolean getDoubleDropsDisabled() {
        return Config.getInstance().getDoubleDropsDisabled(this);
    }

    public ToolType getToolType() {
        return toolType;
    }

    public double getXpModifier() {
        return Config.getInstance().getForumulaMultiplier(this);
    }

    public static SkillType getSkill(String skillName) {
        if (!Config.getInstance().getLocale().equalsIgnoreCase("en_US")) {
            for (SkillType type : values()) {
                if (skillName.equalsIgnoreCase(LocaleLoader.getString(StringUtils.getCapitalized(type.name()) + ".SkillName"))) {
                    return type;
                }
            }
        }

        for (SkillType type : values()) {
            if (type.name().equalsIgnoreCase(skillName)) {
                return type;
            }
        }

        mcMMO.p.getLogger().warning("[Debug] Invalid mcMMO skill (" + skillName + ")");
        return null;
    }

    // TODO: This is a little "hacky", we probably need to add something to distinguish child skills in the enum, or to use another enum for them
    public boolean isChildSkill() {
        switch (this) {
        case SMELTING:
            return true;

        default:
            return false;
        }
    }
}
